import logging
import threading
from concurrent.futures import ThreadPoolExecutor

import requests
from .helpers import Range

logger = logging.getLogger(__name__)


class InvalidUrl(Exception):
    pass


def divide(length, segment_size):
    """将一跟线段细分成多段

    """
    segment_count = int(length / segment_size)
    if (length % segment_size) != 0:
        segment_count += 1
    for i in range(0, segment_count):
        if i == segment_count - 1:
            start, end = i * segment_size, length
        else:
            start, end = i * segment_size, segment_size * (i + 1)
        yield (i, start, end)


class FileDownloadTask:
    def __init__(self, url, filename, progress_cb=None):
        self.url = url
        self.filename = filename
        self.seek_write_lock = threading.Lock()

        self.progress_cb = progress_cb

    def run(self, executor, http, segment_size=1024*1024*4):
        self.executor = executor
        self.http = http
        self.segment_size = segment_size

        self._dl()

    def _get_length(self):
        """发送 HEAD 请求，尝试获取文件大小

        如果能获取，则返回长度，否则返回 None，遇到时错误抛出异常。
        """
        resp = self.http.head(self.url, timeout=1)
        status_code = resp.status_code
        if status_code >= 400:
            logger.warning('Get file length failed, status:%d', status_code)
            if status_code == 403:
                logger.warning('It seems that the url:%s is expired', self.url)
            raise InvalidUrl("Head request failed, status:{}".format(status_code))
        if 200 <= status_code < 300:
            length = int(resp.headers.get('content-length'))
            return length

    def _dl(self):
        url = self.url
        executor = self.executor

        length = self._get_length()
        if length is None:
            raise RuntimeError("only support downloading those files with length.")

        with open(self.filename, 'wb') as f:
            futures = []
            for _, start, end in divide(length, self.segment_size):
                futures.append(executor.submit(self._dl_range, f, url, start, end, length))
            for future in futures:
                future.result()  # get each future result to detect exception

    def _dl_range(self, f, url, start, end, length):
        http = self.http
        headers = {'Range': Range('bytes', [(start, end)]).to_header()}
        resp = http.get(url, headers=headers, stream=True, timeout=5)
        size = 0
        for chunk in resp.iter_content(1024 * 8):
            with self.seek_write_lock:
                f.seek(start + size)
                f.write(chunk)
            size += len(chunk)
            if self.progress_cb is not None:
                try:
                    self.progress_cb(start, size + start, end, length)
                except:  # noqa
                    logger.exception('Progress callback failed.')


class Downloader:
    def __init__(self, max_workers=None):
        self.dl_executor = ThreadPoolExecutor(max_workers=max_workers)
        self.dl_range_executor = ThreadPoolExecutor(max_workers=10)

    def create_task(self, url, filename, http=None, progress_cb=None):
        task = FileDownloadTask(url, filename, progress_cb=progress_cb)
        return self.dl_executor.submit(
            task.run, self.dl_range_executor, http or requests)



from .downloader import Downloader
from .progress import ConsoleProgress


def download(url, filename, console=False):
    dler = Downloader()
    progress_cb = None
    if console is True:
        progress_cb = ConsoleProgress().on_update
    return dler.create_task(url, filename, progress_cb=progress_cb)



import asyncio
import os
import logging

logger = logging.getLogger(__name__)


class Downloader:
    async def run(self, url, filepath, **kwargs):
        pass

    def clean(self, filepath):
        if os.path.exists(filepath):
            try:
                os.remove(filepath)
            except:  # noqa
                logger.exception(f'remove {filepath} failed, this should not happen')
            else:
                logger.info('download file faield, remove the tmp file')


class CurlDownloader(Downloader):
    async def run(self, url, filepath, **kwargs):
        try:
            proc = await asyncio.create_subprocess_exec(
                'curl', url, '--output', filepath)
        except FileNotFoundError:
            logger.error("can't start download, curl is not installed")
            raise
        else:
            await proc.wait()
        return proc.returncode == 0



"""
fuo_dl
------
fuo_dl 是 FeelUOwn 的一个音乐下载插件。

fuo_dl 支持多首歌并行下载，也支持一首歌分多段并行下载（也就是常说的多线程下载）。
另外，用户可以在 ~/.fuorc 中自定义下载路径::

   config.DOWNLOAD_DIR = '~/Music'
"""

import logging

from .manager import DownloadManager

__alias__ = '音乐下载'
__desc__ = __doc__
__version__ = '0.1'


logger = logging.getLogger(__name__)


dm_mgr = None
dm_ui = None


def enable(app):
    global dm_mgr, dm_ui

    # initialize download manager
    dm_mgr = dm_mgr or DownloadManager(app)
    dm_mgr.initialize()

    # initialize ui for download manager
    if app.mode & app.GuiMode:
        from .ui import DownloadUi  # noqa

        dm_ui = dm_ui or DownloadUi(dm_mgr, app, app.ui)
        dm_ui.initialize()


def disable(app):
    pass



import logging

from fuocore import aio
from fuocore.models import ModelType
from feeluown.helpers import async_run
from feeluown.widgets.statusline import StatusLineItem

from .manager import DownloadManager
from .helpers import cook_filename
from .statusline import DownloadLabel


logger = logging.getLogger(__name__)


class DownloadUi:
    def __init__(self, mgr: DownloadManager, app, ui):
        """

        :type app: feeluown.app.App
        :type ui: feeluown.ui.Ui
        """
        self._mgr = mgr
        self._app = app
        self._ui = ui
        self.cur_song_dl_btn = ui.pc_panel.download_btn
        self._ui.bottom_panel.status_line.add_item(
            StatusLineItem('download', DownloadLabel(self._app, mgr)))

    def initialize(self):
        self._app.playlist.song_changed.connect(self._on_song_changed, aioqueue=True)

        # bind download btn clicked signal
        try:
            self.cur_song_dl_btn.disconnect()
        except TypeError:  # signal has no slots
            pass
        self.cur_song_dl_btn.clicked.connect(self._download_cur_media)
        self._ui.songs_table.about_to_show_menu.connect(self._add_download_action)

    def _on_song_changed(self, song):

        async def func(song):
            if song is None:
                self.cur_song_dl_btn.setEnabled(False)
                return

            title = await async_run(lambda: song.title)
            artists_name = await async_run(lambda: song.artists_name)

            filename = cook_filename(title, artists_name)
            is_downloaded = self._mgr.is_file_downloaded(filename)
            if is_downloaded:
                self.cur_song_dl_btn.setEnabled(False)
                self.cur_song_dl_btn.setChecked(True)
            else:
                self.cur_song_dl_btn.setEnabled(True)
                self.cur_song_dl_btn.setChecked(False)

        aio.create_task(func(song))

    def _download_cur_media(self):

        song = self._app.playlist.current_song
        if song is None:
            return
        aio.create_task(self.download_song(song))

    async def download_song(self, song):
        title = await async_run(lambda: song.title)
        artists_name = await async_run(lambda: song.artists_name)

        filename = cook_filename(title, artists_name)
        if self._mgr.is_file_downloaded(filename):
            return

        media = self._app.player.current_media
        if media.url:
            await self._mgr.get(media.url, filename)
        else:
            # this should not happen, so we log a error msg
            logger.error('url of current song is empty, will not download')

    def _add_download_action(self, ctx):
        def download(models):
            for model in models:
                if model.meta.model_type == ModelType.song:
                    logger.info(f'add download task: {model}')
                    aio.create_task(self.download_song(model))

        add_action = ctx['add_action']
        add_action('下载歌曲', download)



from enum import Enum

from .base_downloader import Downloader


class DownloadStatus(Enum):
    pending = 'pending'
    running = 'running'
    ok = 'ok'
    failed = 'failed'


class DownloadTask:
    def __init__(self, url, filename, downloader):
        self.url = url
        self.filename = filename

        self.downloader: Downloader = downloader
        #: task status
        self.status: DownloadStatus = DownloadStatus.pending



"""
一些辅助类

.. note::

     ContentRange, Range, parse_range_header, parse_content_range_header
     四个实现均是从 werkzeug 项目中拷贝过来。
"""


class ContentRange(object):

    """Represents the content range header.

    .. versionadded:: 0.7
    """

    def __init__(self, units, start, stop, length=None, on_update=None):
        assert is_byte_range_valid(start, stop, length), \
            'Bad range provided'
        self.on_update = on_update
        self.set(start, stop, length, units)

    def _callback_property(name):
        def fget(self):
            return getattr(self, name)

        def fset(self, value):
            setattr(self, name, value)
            if self.on_update is not None:
                self.on_update(self)
        return property(fget, fset)

    #: The units to use, usually "bytes"
    units = _callback_property('_units')
    #: The start point of the range or `None`.
    start = _callback_property('_start')
    #: The stop point of the range (non-inclusive) or `None`.  Can only be
    #: `None` if also start is `None`.
    stop = _callback_property('_stop')
    #: The length of the range or `None`.
    length = _callback_property('_length')

    def set(self, start, stop, length=None, units='bytes'):
        """Simple method to update the ranges."""
        assert is_byte_range_valid(start, stop, length), \
            'Bad range provided'
        self._units = units
        self._start = start
        self._stop = stop
        self._length = length
        if self.on_update is not None:
            self.on_update(self)

    def unset(self):
        """Sets the units to `None` which indicates that the header should
        no longer be used.
        """
        self.set(None, None, units=None)

    def to_header(self):
        if self.units is None:
            return ''
        if self.length is None:
            length = '*'
        else:
            length = self.length
        if self.start is None:
            return '%s */%s' % (self.units, length)
        return '%s %s-%s/%s' % (
            self.units,
            self.start,
            self.stop - 1,
            length
        )

    def __nonzero__(self):
        return self.units is not None

    __bool__ = __nonzero__

    def __str__(self):
        return self.to_header()

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, str(self))


class Range(object):

    """Represents a range header.  All the methods are only supporting bytes
    as unit.  It does store multiple ranges but :meth:`range_for_length` will
    only work if only one range is provided.

    .. versionadded:: 0.7
    """

    def __init__(self, units, ranges):
        #: The units of this range.  Usually "bytes".
        self.units = units
        #: A list of ``(begin, end)`` tuples for the range header provided.
        #: The ranges are non-inclusive.
        self.ranges = ranges

    def range_for_length(self, length):
        """If the range is for bytes, the length is not None and there is
        exactly one range and it is satisfiable it returns a ``(start, stop)``
        tuple, otherwise `None`.
        """
        if self.units != 'bytes' or length is None or len(self.ranges) != 1:
            return None
        start, end = self.ranges[0]
        if end is None:
            end = length
            if start < 0:
                start += length
        if is_byte_range_valid(start, end, length):
            return start, min(end, length)

    def make_content_range(self, length):
        """Creates a :class:`~werkzeug.datastructures.ContentRange` object
        from the current range and given content length.
        """
        rng = self.range_for_length(length)
        if rng is not None:
            return ContentRange(self.units, rng[0], rng[1], length)

    def to_header(self):
        """Converts the object back into an HTTP header."""
        ranges = []
        for begin, end in self.ranges:
            if end is None:
                ranges.append(begin >= 0 and '%s-' % begin or str(begin))
            else:
                ranges.append('%s-%s' % (begin, end - 1))
        return '%s=%s' % (self.units, ','.join(ranges))

    def to_content_range_header(self, length):
        """Converts the object into `Content-Range` HTTP header,
        based on given length
        """
        range_for_length = self.range_for_length(length)
        if range_for_length is not None:
            return '%s %d-%d/%d' % (self.units,
                                    range_for_length[0],
                                    range_for_length[1] - 1, length)
        return None

    def __str__(self):
        return self.to_header()

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, str(self))


def is_byte_range_valid(start, stop, length):
    """Checks if a given byte content range is valid for the given length.

    .. versionadded:: 0.7
    """
    if (start is None) != (stop is None):
        return False
    elif start is None:
        return length is None or length >= 0
    elif length is None:
        return 0 <= start < stop
    elif start >= stop:
        return False
    return 0 <= start < length


def parse_range_header(value, make_inclusive=True):
    """Parses a range header into a :class:`~werkzeug.datastructures.Range`
    object.  If the header is missing or malformed `None` is returned.
    `ranges` is a list of ``(start, stop)`` tuples where the ranges are
    non-inclusive.

    .. versionadded:: 0.7
    """
    if not value or '=' not in value:
        return None

    ranges = []
    last_end = 0
    units, rng = value.split('=', 1)
    units = units.strip().lower()

    for item in rng.split(','):
        item = item.strip()
        if '-' not in item:
            return None
        if item.startswith('-'):
            if last_end < 0:
                return None
            try:
                begin = int(item)
            except ValueError:
                return None
            end = None
            last_end = -1
        elif '-' in item:
            begin, end = item.split('-', 1)
            begin = begin.strip()
            end = end.strip()
            if not begin.isdigit():
                return None
            begin = int(begin)
            if begin < last_end or last_end < 0:
                return None
            if end:
                if not end.isdigit():
                    return None
                end = int(end) + 1
                if begin >= end:
                    return None
            else:
                end = None
            last_end = end
        ranges.append((begin, end))

    return Range(units, ranges)


def parse_content_range_header(value, on_update=None):
    """Parses a range header into a
    :class:`~werkzeug.datastructures.ContentRange` object or `None` if
    parsing is not possible.

    .. versionadded:: 0.7

    :param value: a content range header to be parsed.
    :param on_update: an optional callable that is called every time a value
                      on the :class:`~werkzeug.datastructures.ContentRange`
                      object is changed.
    """
    if value is None:
        return None
    try:
        units, rangedef = (value or '').strip().split(None, 1)
    except ValueError:
        return None

    if '/' not in rangedef:
        return None
    rng, length = rangedef.split('/', 1)
    if length == '*':
        length = None
    elif length.isdigit():
        length = int(length)
    else:
        return None

    if rng == '*':
        return ContentRange(units, None, None, length, on_update=on_update)
    elif '-' not in rng:
        return None

    start, stop = rng.split('-', 1)
    try:
        start = int(start)
        stop = int(stop) + 1
    except ValueError:
        return None

    if is_byte_range_valid(start, stop, length):
        return ContentRange(units, start, stop, length, on_update=on_update)


def cook_filename(title, artists_name):
    return f'{title} - {artists_name}.mp3'



from PyQt5.QtCore import Qt, QRect, QTimer
from PyQt5.QtGui import QColor, QPalette

from feeluown.widgets.statusline import StatuslineLabel


class DownloadLabel(StatuslineLabel):
    def __init__(self, app, mgr):
        super().__init__(app, parent=None)

        mgr.tasks_changed.connect(self._on_tasks_changed)

        self._timer = QTimer(self)
        self._downloading = False
        self._status_color = 'orange'

        self._timer.timeout.connect(self._on_timer_timeout)
        self._timer.start(1000)

    def _on_timer_timeout(self):
        self._schedule_update()

    def _on_tasks_changed(self, tasks):
        self.setText(str(len(tasks)))
        if len(tasks) > 0:
            self._downloading = True
            self._timer.start()
        else:
            self._downloading = False
            self._timer.stop()
            self._schedule_update()

    def _schedule_update(self):
        """schedule an paint event"""
        self.update()

    def drawInner(self, painter):
        fill_color = self.palette().color(QPalette.Text)
        painter.setBrush(QColor(fill_color))
        painter.setPen(Qt.NoPen)

        # parameters
        inner_width, inner_height = self._inner_width, self._inner_height
        line_width, line_width_half, line_radius = 2, 1, 1

        # draw an arrow in inner rect
        center = inner_width // 2

        # draw the vertical line
        painter.save()
        line_height = 9
        point_height = line_width
        point_top_margin, point_line_margin = 1, 2
        line_top_margin = point_top_margin + point_height + point_line_margin
        x = center - line_width_half
        point_rect = QRect(x, point_top_margin, line_width, line_width)
        line_rect = QRect(x, line_top_margin, line_width, line_height)
        painter.drawRoundedRect(line_rect, line_radius, line_radius)
        painter.drawRoundedRect(point_rect, line_radius, line_radius)

        # draw left and right line
        vertical_line_bottom_y = line_rect.y() + line_rect.height()
        side_line_width, side_line_height = line_width, 6
        angle = 40
        translate_x = inner_width // 2
        translate_y = vertical_line_bottom_y
        side_line_rect = QRect(-side_line_width//2, -side_line_height,
                               side_line_width, side_line_height)
        painter.translate(translate_x, translate_y)
        painter.rotate(-angle)
        painter.drawRoundedRect(side_line_rect, line_radius, line_radius)
        painter.rotate(angle * 2)
        painter.drawRoundedRect(side_line_rect, line_radius, line_radius)
        painter.restore()

        # draw bottom line
        margin_left = 2
        bottom_line_height = 2
        bottom_lr_height = 4
        bottom_line_y = inner_height - 2
        bottom_line_width = inner_width - margin_left * 2
        bottom_line_rect = QRect(margin_left, bottom_line_y,
                                 bottom_line_width, bottom_line_height)
        bottom_lr_y = bottom_line_y - bottom_lr_height + bottom_line_height
        bottom_left_rect = QRect(margin_left, bottom_lr_y,
                                 line_width, bottom_lr_height)
        bottom_right_rect = QRect(margin_left+bottom_line_width-line_width, bottom_lr_y,
                                  line_width, bottom_lr_height)
        painter.drawRoundedRect(bottom_line_rect, 2, 2)
        painter.drawRoundedRect(bottom_left_rect, 2, 2)
        painter.drawRoundedRect(bottom_right_rect, 2, 2)

    def drawStatus(self, painter):
        if self._downloading:
            super().drawStatus(painter)


if __name__ == '__main__':
    from unittest.mock import MagicMock
    from PyQt5.QtWidgets import QApplication

    app = QApplication([])
    widget = DownloadLabel(MagicMock(), MagicMock())
    widget.show()

    widget.resize(30, 30)
    app.exec()



import sys
import threading


class ConsoleProgress:
    def __init__(self):
        self.lock = threading.Lock()
        self.progress = {}

    def on_update(self, start, current, end, length):
        with self.lock:
            progress = self.progress
            progress[(start, end)] = current

            fill = "█"
            not_fill = "-"
            total = 50

            finished = 0
            fill_pos = set()
            s_points = set()
            e_points = set()
            print('\rDownload Progress: ', end='')
            for r, c in progress.items():
                s, e = r
                finished += c - s
                s_points.add(s)
                e_points.add(e)
                i1 = int(s / length * total)
                i2 = int(c / length * total)
                while i1 <= i2:
                    fill_pos.add(i1)
                    i1 += 1
            for num in range(0, total + 1):
                c = fill if num in fill_pos else not_fill
                print(c, end='')
            percent = int(finished / length * 100)
            print(' {}%'.format(percent), end='')
            sys.stdout.flush()
            if finished == length and s_points ^ e_points == {0, length}:
                print('\n')



import asyncio
import logging
import os

from typing import List

from fuocore import aio
from fuocore.dispatch import Signal
from feeluown.consts import SONG_DIR
from .task import DownloadTask, DownloadStatus
from .base_downloader import Downloader, CurlDownloader

logger = logging.getLogger(__name__)


class DownloadManager:
    def __init__(self, app):
        """

        :type app: feeluown.app.App
        """
        self._tasks = []
        self._task_queue = asyncio.Queue()

        #: emit List[DownloadTask]
        self.tasks_changed = Signal()
        self.downloader: Downloader = CurlDownloader()

        self._path = SONG_DIR

    def initialize(self):
        aio.create_task(self.worker())

    def list_tasks(self) -> List[DownloadTask]:
        return self._tasks

    async def get(self, url, filename, headers=None, cookies=None):
        """download and save a file

        :param url: file url
        :param filename: target file name
        """
        # check if there exists same task
        for task in self.list_tasks():
            if task.filename == filename:
                logger.warning(f"task:{filename} has already been put into queue")
                return

        task = DownloadTask(url, filename, self.downloader)
        self._tasks.append(task)
        await self._task_queue.put(task)
        self.tasks_changed.emit(self.list_tasks())
        logger.info(f"task:{filename} has been put into queue")

    async def worker(self):
        while True:
            task = await self._task_queue.get()
            await self.run_task(task)

            # task done and remove task
            self._task_queue.task_done()
            self._tasks.remove(task)
            self.tasks_changed.emit(self.list_tasks())

            path = self._getpath(task.filename)
            logger.info(f'content has been saved into {path}')

    async def run_task(self, task):
        task.status = DownloadStatus.running
        filename = task.filename
        downloader = task.downloader

        filepath = self._getpath(filename)
        try:
            ok = await downloader.run(task.url, filepath)
        except asyncio.CancelledError:
            task.status = DownloadStatus.failed
        else:
            if ok:
                task.status = DownloadStatus.ok
            else:
                task.status = DownloadStatus.failed

        # clean up the temp file if needed
        if task.status is DownloadStatus.failed:
            downloader.clean(filepath)

    def is_file_downloaded(self, filename):
        return os.path.exists(self._getpath(filename))

    def _getpath(self, filename):
        return os.path.join(self._path, filename)



#!/usr/bin/env python3

from setuptools import setup


setup(
    name='fuo_dl',
    version='0.1',
    description='feeluown download plugin',
    author='Cosven',
    author_email='yinshaowen241@gmail.com',
    packages=[
        'fuo_dl',
    ],
    package_data={
        '': []
        },
    url='https://github.com/feeluown/feeluown-download',
    keywords=['feeluown', 'plugin', 'download'],
    classifiers=(
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only',
        ),
    install_requires=['feeluown>=3.5.3', 'requests'],
    entry_points={
        'fuo.plugins_v1': [
            'download = fuo_dl',
        ]
    },
)
